/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   mo_fedotov@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/wait.h>

#include <string.h>

#include <qdatetime.h>
#include <qfile.h>
#include <qtimer.h>

#include <kuniqueapplication.h>

#include "ar_client.h"
#include "ar_parser.h"
#include "ar_cmds.h"
#include "bt_comm.h"
#include "logger.h"

extern logger2*     	    theLog;
extern AR_Parser*   	    conf;
extern AR_Cmds*				commander;
extern KUniqueApplication*  pApp;
extern KAnyRemote*  	    mainWin;

AR_Client::AR_Client()
{
    //theLog->debug("AR_Client::AR_Client");
    callerId 	    = QString("NO CALLER ID");
    callOn          = false;
    modifierString  = "";
    repeatCmdPtr    = NULL;	
    repeatNow       = false;
    useCallId	    = true;
    serverMode      = SERVER_NO;
    initialized     = 0;    
    serverMode	    = 0;
	//paramValue		= -1;
}


AR_Client::~AR_Client()
{
}

bool AR_Client::isConnected()
{
    return Comm.isConnected();
}

bool AR_Client::isServer()
{
    return (serverMode > 0);
}

type_key*  AR_Client::getAutoRepeatCmdIndex()
{
    return repeatCmdPtr;
}


void AR_Client::setStyle(QString device)    	// Determine is we were run as server or as client
{   	    	    	    	    	 
    // device should be in format:
    // 1. /dev/something
    // 2. socket:5000
    // 3. bluetooth:10 or bluetooth:default (port will be 10 by default)
    
    if (device.contains(INET_SOCKET)) {
        port = (device.mid(QString(INET_SOCKET).length())).toInt();
        theLog->debug("TCP Server mode. Will use port " + QString().setNum(port));
    	serverMode = SERVER_TCP;
    } else if (device.contains(BT_SOCKET)) {
        port = 10;
        QString chan("10");
        if (!device.contains(":")) {
            chan = device.mid(QString(BT_SOCKET).length()+1);
            port = chan.toInt();
        }
        theLog->debug("Bluetooth Sockets Server mode. Will use channel " + chan);
    	serverMode = SERVER_BT;
    } else {
    	serverMode = SERVER_NO;
		port = -1;
    }
}

bool AR_Client::initServerPort()    	    // If acts as server then open socket
{   	    	    	    	    	    // if acts as client do nothing
    if (serverMode == SERVER_TCP) {
    	theLog->info("initPort() opening TCP socket");
        if (Comm.openSocketPort(SERVER_TCP, port) < 0) {
            return false;
        }
    } else if (serverMode == SERVER_BT) {
    	theLog->info("initPort() opening BT socket");
    	if (Comm.openSocketPort(SERVER_BT, port) < 0) {
            return false;
        }
    } else {
    	theLog->info("initPort() nothing to do");
    }
    
    return true;
}

void AR_Client::setInitialized() {
    if (initialized == 0) { 
    	// Run "onInit" command
    	int i = 0;
    	type_key *k = conf->findItem(QString(EVT_INIT), i);
    	if (k && i == 1) {
    	    theLog->info("Exec cmd on init");
    	    commander->handleCmdByKey(k);  // Since we not connected yet we did not interested in return code
    	} else {
    	    theLog->info("No command on init\n");
    	}
    	initialized++;
    }
}

int AR_Client::listenForConnection() {
   
    if (serverMode == SERVER_NO) {
    	return -1;
    }   

    int ret = Comm.listenAndAcceptSocketConn(serverMode);
    
    if (ret == 1) { 	    	// Connection established
    	int i = 0;
    	type_key* tk = conf->findItem(EVT_CONNECT, i);
    	if (tk != NULL && i == 1) {
    	    commander->handleCmdByKey(tk);
    	} 
    }
    return ret;
}

// Used only in client style
bool AR_Client::connectToBT(QString port)
{
    //theLog->debug("AR_Client::connectToBT");
    
    if (Comm.initPort(port)) {
    	theLog->info("Modem initializing");
 
    	QString answer;
    	Comm.sendCommand("AT\r", answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString(" AT -> ") + answer);
    	    Disconnect();
    	    return false;
    	}

    	Comm.sendCommand("ATE0\r", answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString("ATE0 -> ") + answer);
    	    Disconnect();
    	    return false;
    	}

    	QString chrset = conf->getCharset();
    	if (!(chrset.isNull() || chrset.isEmpty())) {
    	    Comm.sendCommand(QString(DEF_AT_CSCS) + "\"" + chrset + "\"\r",answer, QString::null);
    	    if (answer.contains("OK") == 0) {
    		theLog->error(QString("Can't set charset to ") + chrset);
    	    }		
    	}
    	pApp->processEvents();

    	Comm.sendCommand(QString(DEF_AT_CLIP) + "\r",answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString("Can't set CLIP ON"));
    	}
    	pApp->processEvents();
	
    	// Determine model
    	theLog->debug("Determine phone model");
    	Comm.sendCommand(QString(DEF_AT_CGMI) + "\r", answer, QString::null);
    	
	
    	//Set model in conf. data
		conf->setModel(answer);
	    
    	QString cmer = conf->getAT_CMER(CMER_ON) + "\r";

    	Comm.sendCommand(cmer, answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString("AT_CMER SET >") + cmer + "< ->" + answer);
    	    Disconnect();
    	    return false;
    	}
    	pApp->processEvents();

    	// Siemens S55 needs additional AT+CMEC=2 to make AT+CKPD works
    	// not sure about other Siemens phones
    	if (conf->getModel() == MODEL_SIEMENS) {
    	    Comm.sendCommand(QString(DEF_AT_CMEC) + "\r", answer, QString::null);
    	    if (answer.contains("OK") == 0) {
    		theLog->error(QString("Failed in AT+CMEC"));
    	    }
    	}

    	// Will set global callerId
    	int ret = Comm.getClip(callerId);
    	if(ret == EXIT_NOK) {		// Got ERROR on AT+CLCC; probably phone does not supports this command
    	    theLog->error(QString("fails in getClip"));
    	    useCallId = false;
    	}
    }
    
    setInitialized();

    manageAlarms(ALARM_CLEAN);

    if (serverMode == SERVER_NO) {
    	// These is not fatal errors
    	if (Comm.sendSeq(conf->getToMainMenu()) != EXIT_OK) {
    	    theLog->debug("Fails in sendSeq(getToMainMenu())");
    	}
    }
    pApp->processEvents();

    theLog->info("Connection established");
    //setRemoteOn(true);

    int i = 0;
    type_key* tk = conf->findItem(EVT_CONNECT, i);
    if (tk != NULL && i == 1) {

    	int ret = commander->handleCmdByKey(tk);
    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    return true;    // Return true because connection already established
    	}
    } else {
    	theLog->info("Start event forwarding");
    }

    return true;
}

// This is not clean-up function. It just finished the connection.
void AR_Client::Disconnect()
{
    theLog->info("Finish events forwarding");

    pApp->processEvents();

    if (serverMode == 0) {
        QString answer;
        QString cmer = conf->getAT_CMER(CMER_OFF) + "\r";
    
        Comm.sendCommand(cmer, answer, QString::null);
        //handleCommand((char *)answer.ascii()); // Does it needed ?
    
        // and resend No back
        //Comm.sendCommand("AT+CKPD=\"e\"\r", answer);  seems it is not obligatory
    } 
    
    pApp->processEvents();
    Comm.closePort(1);
    
    return;
}

void AR_Client::handleDisconnect(bool needClose)
{
    theLog->debug("handleDisconnect");

    int i;
    type_key *tk = conf->findItem(EVT_DISCONNECT, i);
    if (tk != NULL) {
    	i = commander->handleCmdByKey(tk);
	// No additional handling since we already disconnected
    }
    
    if (needClose) {	    // in server style we wan't do this
    	Comm.closePort(1);
    }
    return;
}

int AR_Client::handleKeyPress(char* keyIn) 
{
    int ret;
    
    theLog->debug(QString("AR_Client::handle_key_press() >") + keyIn + "<");
    
    // Search an alias
    QString key;
    if((key = conf->findAlias(keyIn)).isNull()) {
        key = keyIn;
    }

    // If modifier was set already it needs to verify which multi-key command we collect
    if (!modifierString.isEmpty()) {
    	modifierString += " "; // Add one space first
    	modifierString += key;
    	theLog->debug("Modifier+sequence is: >" + modifierString + "<");
    }
    
    // Execute appropriate command
    int i = 1;
    type_key *tk = conf->findItem(modifierString.isEmpty() ? QString(key) : modifierString, i); 
    if (tk != NULL && i == 1) {
    	theLog->debug("One-key command found");
	
    	if ((!repeatNow) && conf->getAutoRepeat()) {
 	    theLog->debug("Set auto repeat flag");
 	    repeatNow    = true;
 	    repeatCmdPtr = tk;
     	} 
	
	ret = commander->handleCmdByKey(tk);
	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    return ret;
    	}
	
	// AR_Client is not successor of QObject so send timeout to mainWin
	if (conf->getAutoRepeat()) {
	    theLog->debug("Set auto repeat timer");
	    QTimer::singleShot(1000, mainWin, SLOT(autoRepeat()));
	}
	    
    	// Clean modifier string
    	modifierString = "";

    } else if (tk != NULL && i == 0) {  // A part of multi-key command
        theLog->debug("Part of multi key command found");
	
    	if (modifierString.isEmpty()) {
	    theLog->debug("Start of multi key sequence");
    	    modifierString = key;
    	} 
	    
        if(serverMode == SERVER_NO) {
            ret = Comm.sendSeq(conf->getToMainMenu());
	    if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	        return ret;
    	    }
        }

    } else {
	// User had pressed some key ... 
    	// Send ToMainMenu sequence of CKPD's to show main menu again
    	if (serverMode == SERVER_NO && Comm.sendSeq(conf->getToMainMenu()) != EXIT_OK) {
    	    theLog->debug("Fails in sendSeq()");
    	}

        theLog->debug("No approprite key definition was found by findItem()");
	
    	if (!modifierString.isEmpty()) {
	    // Clean modifier string
    	    modifierString = "";
	    theLog->debug("Clean modifier string");
    	}
    }

    return EXIT_OK;
}

// Should be called to handle inpit from BT device
int AR_Client::handleCommand(char* cmd) 
{
    char *prev, *next, *key;
    int ret, i;

    commander->setParamValue(-1);	// clean up param value
    
    //skip lines starting with \n and \r
    if (cmd[0] == '\r') cmd++;
    if (cmd[0] == '\n') cmd++;

    theLog->debug(QString("AR_Client::handle_command() >") + cmd + "<");
  
    // most common case
    if (!cmd[0]) {
    	return EXIT_OK;
    }
    
    // if received multiline command - handle line by line and return
    prev = cmd;
    next = strchr(cmd, '\r');
    if (next == NULL && serverMode > 0) {   // Java client wiil send +CKEV: 1,1; +CKEV: 1,0
    	next = strchr(cmd, '\r');
    }
    if (next) {
    	theLog->debug("handle_command() multiline");
	
        char copy[1024];
        int len;
    
        do {     
            len = next-prev;
            
            if (len > 2) {
                memcpy(copy, prev, len);
                copy[len] = 0;
                
                // use recursion
                ret = handleCommand(copy);
	    	if (ret == EXIT_DISCON || ret == EXIT_ABORT) {
	    	    return ret;
	    	}
            }
            prev = next+1;
        } while ((next = strchr(prev, '\r'))) ;
    
        // now return
        return EXIT_OK;
    }
    if (strstr(cmd, "OK") != NULL) {	// OK - nothing to do
        return EXIT_OK;
    }
    
    // Handle different keypresses
    // Motorola: +CKEV: "1",1
    // SE      : +CKEV: 1,1
    // Sagem: in rare ) cases it could be (between > and <) >,1:< ?
    if (memcmp(cmd, "+CKEV:", 6) == 0) {
    	theLog->debug("Got +CKEV");
	
	int isSE = MODEL_SE;
	
        key = cmd+7;

	// Sagem test
	//key+=3;

	if (*key == '"') {	// Seems this is Motorola
	    key++;
	    isSE = MODEL_MOTOROLA;
	}

	if(*key == ',') {	// Seems this is Sagem
	    theLog->debug("+CKEV is in rare Sagem format (Empty code)");
	    isSE = MODEL_SAGEM;
	}

	i = 1;
	if (isSE == MODEL_SE) {	// SonyEricsson
	    while(key[i] != ',') i++;
	    key[i] ='\0';
	    
	    i = i+1;
	} else if (isSE == MODEL_MOTOROLA) {	// Motorola
	    while(key[i] != '"') i++;
	    key[i] ='\0';

	    i = i+2;
	} else if (isSE == MODEL_SAGEM) {	// rare Sagem case
	    *key ='\0';
     	} else {
	    theLog->error("Can't recognize +CKEV event !!!");
	    return EXIT_NOK;
	}
    	
	if (key[i] == '1') {
	    theLog->debug("Button down event");
            return handleKeyPress(key);
    	} else if (key[i] == '0') {
	    theLog->debug("Button up event");
            // In general we skip this
	    if (repeatNow) {
	    	// Clear auto repeat flag
		theLog->debug("Clean auto repeat flag");
    	    	repeatNow    = false;
    	    	repeatCmdPtr = NULL;
	    }
	    return EXIT_OK;
    	} else {
    	    theLog->debug(QString("something wrong: key should be pressed or released: cmd=>") + key + "<");
            return EXIT_NOK;
    	}
    } else if (memcmp(cmd, DEF_CLCC,6) == 0) {
	theLog->debug("Caller ID received. Skip it.");
	return EXIT_OK;
    } else if (strstr(cmd, DEF_RING) != NULL) {     // Incoming call
	theLog->info("Incoming call");     	    // This event sent periodically until used answered a call
	
	if (!callOn) {
	    if (useCallId) {
	    	theLog->debug("Get CallerId on incoming call");
    	    	ret = Comm.getClip(callerId);

		theLog->debug("CallerId is " + callerId);
		
	    	if (ret == EXIT_EXACT) { // Active call exists
		    theLog->debug("Caller ID received");
	    	} else { 
		    theLog->debug("Can't get Caller ID. Probably call was finished");
		}

	    }
	    //
	    // May be we can get +CLIP: "+xxxxxxxxxxxx",145,,,,0< here ?
	    // 
            
            i = 0;
	    type_key *tk = conf->findItem(EVT_INCALL, i);
	
    	    if (tk != NULL && i == 1) {
	    	theLog->info("Exec cmd on incoming call");
  	    	ret = commander->handleCmdByKey(tk);
	    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    return ret;
    	    	}
	    }
	
	    	
	    // If there is (EndCall) command,then set timer to determine when call ends
	    i = 0;
            tk = conf->findItem(EVT_ENDCALL, i);
	    if (tk != NULL && i == 1) {
	    	theLog->debug("Set call timer");
	    	callOn    = true;
		    
    	    	// AR_Client is not successor of QObject sosend timeout to mainWin
    	    	theLog->debug("Set CallerID timer");
    	    	QTimer::singleShot(1000, mainWin, SLOT(getCallerID()));
	    }
	    
	    return EXIT_OK;
    	} else {
			theLog->debug("Skip it.");      // This event sent periodically until used answered a call
    	}
    } else if (isBemusedCommand(cmd)) {    // Bemused server emulation tricks
    	return EXIT_OK;
    }
    
    theLog->debug(QString("Unhandled cmd ") + cmd);
    return EXIT_OK;
}

int AR_Client::readCommand(QString& buffer, int l) 
{
    int ch, len = 0;
    //theLog->debug("AR_Client::readCommand()");
    
    if (serverMode) {
        len = Comm.readSocketConn(buffer);
        if (len == EOF) {
    	    theLog->debug("AR_Client::readCommand() EOF");
            return EOF;
	}
    	//theLog->debug(QString("AR_Client::readCommand() >") + buffer + "<");
    } else {
        while (len < l) {
            ch = Comm.readchar(100);
            if (ch >= 0 && ch != '\r') {
	    	buffer += ch;
                len++;
            } else  if (ch == EOF) {
    	        buffer = QString::null;
                return EOF;
            } else {
                break;
            }
        }
    }

    return len;
}

// mode: 0 - clear alarms
//       1 - if fired - process, then clear alarms
void AR_Client::manageAlarms(int mode)
{
    //theLog->debug("AR_Client::manageAlarms");

    QString pf;
    cmdItem *pg;
    int i;
    
    int n = conf->getNumAlarms();
    for(i=0;i<n;i++) {
    	pf = conf->getAlarmInfo(i, &pg);
	
    	// If file exists
	QFile fs(pf);
	if (fs.exists()) {
	    if (mode == ALARM_CLEAN) {
		theLog->debug("Remove file");
	    } else {    	
    		theLog->info(QString("Alarm fired (") + pf + ")");
                
	        commander->processCommands(pg);
	    }
	    fs.remove();
	}
    }
    return;
}


bool AR_Client::stillPressed(void)
{
    return repeatNow;
}

int AR_Client::checkCallActive()
{
    theLog->debug("AR_Client::checkCallActive()");
       
    // Is call still active (timeout about 1 seconds) ?
    if (callOn) {
    	theLog->debug("Get call list again");
    	
    	int ret = Comm.getClip(callerId);	    // Will set global callerId
    	if (ret == EXIT_EXACT) {
    	    theLog->debug("Call still active (" + callerId + ")");
    	    return EXIT_OK;
     	} else {
    	    theLog->debug("Call was finished");
	    callOn  	= 0;
    	    callerId	= QString("NO CALLER ID");
			    
    	    type_key* tk = conf->findItem(EVT_ENDCALL, ret);
    	    if (tk != NULL) {
    	    	theLog->debug("EVT_ENDCALL found");
    	    	ret = commander->handleCmdByKey(tk);
    	    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    return ret;
    	    	}
    	    }
   	}
    }
    return EXIT_NOK;
}

BtComm AR_Client::getComm()
{
	return Comm;
}

QString AR_Client::getCallerID()
{
	return callerId;
}
void AR_Client::cleanupRepeatKeys() 
{
	repeatNow    = 0;
	repeatCmdPtr = NULL;
}

//
// Bemused server limited emulation
//
bool  AR_Client::isBemusedCommand(char *cmdIn)
{  
    theLog->debug("isBemusedCommand()");
    
    bool isBemused = false;
    
    char *cmd = cmdIn;
    char *last = cmdIn + strlen(cmdIn);
    
    while (cmd < last) {
    	int handle = 0;
	int shift  = 4;
	char oneCmd[5];
	
    	if (strncmp(cmd, "CHCK", 4) == 0 ||
			strncmp(cmd, "DINF", 4) == 0 ||
    	    strncmp(cmd, "EXIT", 4) == 0 ||
    	    strncmp(cmd, "FADE", 4) == 0 ||
    	    strncmp(cmd, "FFWD", 4) == 0 ||
			strncmp(cmd, "GVOL", 4) == 0 ||
			strncmp(cmd, "INF2", 4) == 0 ||
			strncmp(cmd, "INFO", 4) == 0 ||
			strncmp(cmd, "LIST", 4) == 0 ||
    	    strncmp(cmd, "NEXT", 4) == 0 ||
    	    strncmp(cmd, "PAUS", 4) == 0 ||
			strncmp(cmd, "PLEN", 4) == 0 ||
			strncmp(cmd, "PLST", 4) == 0 ||
    	    strncmp(cmd, "PREV", 4) == 0 ||
			strncmp(cmd, "RMAL", 4) == 0 ||
    	    strncmp(cmd, "RWND", 4) == 0 ||
			strncmp(cmd, "SHUT", 4) == 0 ||
    	    strncmp(cmd, "STEN", 4) == 0 ||
    	    strncmp(cmd, "STOP", 4) == 0 ||
            strncmp(cmd, "STRT", 4) == 0 ||
			strncmp(cmd, "VERS", 4) == 0) {
			handle = 1;
    	}

    	if (strncmp(cmd, "REPT", 4) == 0 ||
            strncmp(cmd, "SHFL", 4) == 0) {
            unsigned char bc = *((unsigned char*) (cmd + 4));
            commander->setParamValue((int)bc);

            handle = 1;
			shift  = 5;
    	}
    	if (strncmp(cmd, "VOLM", 4) == 0) {
			unsigned char bc;
			if (strlen(cmd) == 4) {  // read only VOLM without value to set
				theLog->debug("read one more byte");
				char ret[2];
				ret[1] = '\0';;
				if ((ret[0] = (char) Comm.readchar(500000)) < 0) {
            	    theLog->debug("isBemusedCommand: read < 0");
					commander->setParamValue(-1);
					return 0;
				}
				QString ch(ret);
            	theLog->debug("isBemusedCommand: read >" + QString(ret) + "<");
				bc = (unsigned char) ret[0];
			} else {
				bc = ((unsigned char) *(cmd + 4));
			}
 
    	    // VOLM 0-255 -> 0%-100%
			int p = ((int)bc * 100)/255;
            commander->setParamValue(p);
        
    	    theLog->debug("isBemusedCommand: parameter >" + QString().setNum(p) + "<");	    
	    
			handle = 1;
			shift  = 5;
    	}
    	if (strncmp(cmd, "SLCT", 4) == 0) {
            handle = 1;
			shift  = 6;
    	}
    	if (strncmp(cmd, "SEEK", 4) == 0)  {
			commander->setParamValue((((unsigned char)*(cmd + 4)) << 24) + 
				(((unsigned char)*(cmd + 5)) << 16) + 
				(((unsigned char)*(cmd + 6)) << 8)  + 
				(unsigned char)*(cmd + 7));
            handle = 1;
			shift  = 8;
    	}
    	if (strncmp(cmd, "DLST", 4) == 0 || 
            strncmp(cmd, "DOWN", 4) == 0 || 
            strncmp(cmd, "FINF", 4) == 0 ||    
    	    strncmp(cmd, "LADD", 4) == 0 ||
    	    strncmp(cmd, "PLAY", 4) == 0) {             
	    
            unsigned char bc = ((unsigned char) *(cmd + 4));
	    
            handle = 1;
			shift  = 5 + bc;
    	}
    	if (handle) {
			isBemused = true;
			strncpy(oneCmd,cmd,4);
			oneCmd[4] = '\0';
           
			theLog->debug("isBemusedCommand: one commmand >)" + QString(oneCmd) + "<");	    
    	    
    	    handleKeyPress(oneCmd);
    	} else {
			break;
		}
	
		cmd = cmd + shift;
    }
    return isBemused;
}
