/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   mo_fedotov@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef AR_PARSER_H
#define AR_PARSER_H

/**
	@author M.Fedotov <mo_fedotov@mail.ru>
*/

#include <qstring.h>
#include <qtextstream.h>

#include "kanyremote.h"

#define CFGFILE 	".anyremote.cfg"
#define MAXALARMS    	30
#define DEFAULT_DEVICE  "/dev/rfcomm0"

#define EVT_CONNECT     "(Connect)"
#define EVT_DISCONNECT  "(Disconnect)"
#define EVT_ENDCALL     "(EndCall)"
#define EVT_INCALL     	"(InCall)"
#define EVT_INIT     	"(Init)"
#define EVT_ENTRMODE	"(EnterMode)"
#define EVT_EXITMODE    "(ExitMode)"

#define ID_UNKNOWN    	0
#define ID_EXIT    	1
#define ID_TMPDISCONN   2
#define ID_REMOTEON	3
#define ID_REMOTEOFF    4
#define ID_EXEC         5
#define ID_SENDCKPD     6
#define ID_SETMODE      7
#define ID_SET          8
#define ID_EXECSET      9
#define ID_TIMERCMD     10
#define ID_SEND         11
#define ID_EXECSEND     12
#define ID_CANCELTIMER  13

#define CMD_EXIT    	        "Exit"
#define CMD_TMPDISCONN	        "TempDisconnect"
#define CMD_REMOTEON	        "RemoteOn"
#define CMD_REMOTEOFF	        "RemoteOff"
#define CMD_EXEC                "Exec"
#define CMD_SENDCKPD            "SendCKPD"                  // AT style only
#define CMD_SETMODE	        "SetMode"
#define CMD_SET                 "Set"                       // Server style only
#define CMD_EXECSET             "ExecAndSet"                // Server style only
#define CMD_TIMER               "Timer"                     // Server style only
#define CMD_SEND                "Send"                      // Server style only, mainly for Bemused-server-emulation
#define CMD_EXECSEND            "ExecAndSend"               // Server style only, mainly for Bemused-server-emulation
#define CMD_CANCELTIMER         "CancelTimer"               // Server style only

#define ALARMS_SECTION_STR  "[Alarms]"
#define ALIAS_SECTION_STR   "[Aliases]"
#define KEYS_SECTION_STR    "[Keys]"
#define MODE_END_STR        "[ModeEnd]"
#define MODE_STR    	    "[Mode]"
#define SECTION_END_STR     "[End]"

#define AUTOCONN_STR   	"AutoConnect"
#define AUTOREPEAT_STR  "AutoRepeat"
#define BAUD_STR    	"Baudrate"
#define CHARSET_STR    	"CharSet"
#define CMER_OFF_STR   	"CmerOff"
#define CMER_ON_STR   	"CmerOn"
#define DEBUG_STR   	"Debug"
#define DEVICE_STR   	"Device"
#define LOG_STR    	"Log"
#define RETRY_STR    	"RetrySecs"
#define SCREEN_STR      "Screen"
#define TOMAIN_STR   	"ToMainMenu"
#define EMPTY_STR   	"EMPTY"


// Some AT stuff
// \r will be added before send
#define DEF_AT_CMER_ON_DEFAULT   "AT+CMER=3,2,0,0,0" 	// Default, valid for SE and Siemens S55 phones
#define DEF_AT_CMER_OFF_DEFAULT  "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_SE        "AT+CMER=3,2,0,0,0" 	// SE, same as default
#define DEF_AT_CMER_OFF_SE       "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_MOTOROLA  "AT+CMER=3,1,0,0,0" 	// Motorola phones	
#define DEF_AT_CMER_OFF_MOTOROLA "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_SAGEM     "AT+CMER=2,1,0,0,0" 	// Sagem phones
#define DEF_AT_CMER_OFF_SAGEM    "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_SIEMENS   "AT+CMER=3,2,0,0,0" 	// Siemens phones, same as default
#define DEF_AT_CMER_OFF_SIEMENS  "AT+CMER=0,0,0,0,0" 


#define DEF_AT_CGMI     "AT+CGMI" 
#define DEF_AT_CKPD 	"AT+CKPD="
#define DEF_AT_CLCC     "AT+CLCC"   	    // ? will be added later (will try both variants)
#define DEF_AT_CLIP 	"AT+CLIP=1"
#define DEF_AT_CMEC 	"AT+CMEC=2" 	    // used by Siemens to made AT+CKPD works
#define DEF_AT_CSCS     "AT+CSCS="
#define DEF_CKEV 	"+CKEV:"
#define DEF_CLCC 	"+CLCC:"
#define DEF_RING 	"RING"    	    // It could be RING or +CRING (depends on AT+CRC)

#define CMER_ON  	1
#define CMER_OFF  	2

#define SAME 	    	99
#define PAUSE_STR   	"PAUSE"

#define KEY_NOSPECIAL	0
#define KEY_EXIT    	1
#define KEY_TMPDISCONN	2

#define STR_MOTOROLA	"Motorola"
#define STR_SE   	"Sony Ericsson"
#define STR_SAGEM	"SAGEM"
#define STR_SIEMENS   	"SIEMENS"

#define MODEL_DEFAULT	0
#define MODEL_SE	1
#define MODEL_MOTOROLA	2
#define MODEL_SAGEM	3
#define MODEL_SIEMENS	4

typedef struct Ci
{
    int        type;
    QString    descr;
    struct Ci  *next;
} cmdItem;

typedef struct Tk
{
    QString    key;
    cmdItem    *cmd;
    struct Tk  *next;
} type_key;

typedef struct Ta
{
    QString    key;
    QString    alias;
    struct Ta  *next;
} type_alias;

typedef struct Mt
{
    QString 	name;
    type_key*   keys;
    struct Mt   *next;
} mode;

typedef struct
{
    QString alarmFile;
    cmdItem* cmd;
} type_alarm;

class AR_Parser {
public:
    AR_Parser();
    ~AR_Parser();

public:
	bool	getAutoConnect	(void);
	bool	getAutoRepeat	(void);
	bool	getLog	    	(void);
	bool	getDebug	(void);
	bool	getUseScreen	(void);
	type_key* findItem    	(QString key, int& flag);
	QString   findAlias     (QString keyIn);
	int 	getBaudrate 	(void);
	int 	getNumAlarms  	(void);
	int 	getRetrySecs  	(void);
	int 	load          	(QString file);
	void 	freeCfg         (void);
	QString getAlarmInfo    (int which, cmdItem ** cmds);
	QString getAT_CMER  	(int what);
	QString getCharset  	(void);
	cmdItem* getCommand	(type_key* k);
	QString getDevice       (void);
	QString getInCall  	(void);
	QString getToMainMenu	(void);
	QString parse_CLCC  	(QString clcc);
	void	printConf     	(void);
	void	setAutoConnect 	(bool b);
	void	setDevice   	(QString);
	void	setLog	    	(bool b);
	void	setMode   	(QString mode);
	QString	getMode   	(void);
    	int     getModel      	(void);
    	void    setModel     	(QString m);
	void	truncate    	(void);
	
private:

	void 	loadAlarms	(QTextStream& stream);
	void 	loadAliases	(QTextStream& stream);
	void 	loadKeys	(QTextStream& stream);
	int 	load_cfg	(QString s);
	int  	loadKeys	(FILE *fp);
	int 	setBaudrate	(int rate);
	void 	printAlarms	(void);
	void 	printKeys	(void);
	void 	printParams	(void);
	void 	printAliases	(void);
	void 	storeKey    	(type_key**, QString tag, QString value);
	int     storeCmds   	(cmdItem** cmdList, QString value);
	type_key* findItemInMode(mode *mode, QString keyIn, int& flag);
    	QString id2Cmd      	(int cmdId);
		
	bool autoConnect;
 	bool autoRepeat;
	bool log;
	bool debug;
	bool screen;
 	int  baudrate;
	int  numAlarms;
 	int  retrysecs;
 	QString charset; 	// Charset
	QString cmerOff;
	QString cmerOn;
 	QString device;		// Device
	QString toMain; 	// AnyMenu  -> MainMenu
	type_alarm  AlarmsData[MAXALARMS];
	type_key*   Keys;
	mode*       modes;
	mode*       defMode;
	mode*       currentMode;
	type_alias* aliases;
    	int  	    model;
};

#endif
